# pylint: disable=unused-argument
import importlib
import warnings

warnings.filterwarnings("ignore", category=UserWarning)


def generate_benchmark_v1(
    location: str = "./",
    plot: bool = True,
    pca_plot: bool = True,
    checkpointing: bool = True,
    generate_small: bool = True,
    generate_standard: bool = True,
    generate_dependent: bool = True,
    multi_thread: bool = False,
    max_workers: int = 2
) -> None:
    """
    Generate benchmark datasets with different configurations.

    Parameters
    ----------
    location : str
        The directory where the generated datasets will be stored.
    plot : bool
        Whether to generate plots for the datasets.
    pca_plot : bool
        Whether to generate PCA plots for the datasets.
    checkpointing : bool
        Whether to check for already generated datasets during generation and skip if they are present.
    generate_small : bool
        Whether to generate the set of datasets with a smaller number of data points and simpler configurations.
    generate_standard : bool
        Whether to generate the set of datasets with a standard number of data points and moderate complexity.
    generate_dependent : bool
        Whether to generate the set of datasets with dependent features.
    multi_thread : bool
        Whether to use multi-threading for dataset generation.
    max_workers : int
        The maximum number of worker threads to use if multi_thread is True.

    Raises
    ------
    RuntimeError
        If the installed sid2regenerator version is not 0.1.1 to ensure reproducibility.

    Returns
    -------
    None
    """

    if importlib.metadata.version("sid2regenerator") != "0.1.1":
        raise RuntimeError(
            "This benchmark is based on the sid2re version 0.1.1. "
            f"To ensure reproducibility generating older benchmarks with the currently installed version "
            f"{importlib.metadata.version('sid2regenerator')} is not allowed."
        )



from sid2re.driftgenerator.concept.nodes._base_node import _BaseNode
from sid2re.driftgenerator.concept.nodes.concept_node import RandomConceptFunctionNode
from sid2re.driftgenerator.concept.nodes.function_feature_node import RandomFunctionNode
from sid2re.driftgenerator.concept.nodes.root_feature_node import RootFeatureNode

__all__ = ["RandomConceptFunctionNode", "RandomFunctionNode", "RootFeatureNode", "_BaseNode"]



from sid2re.driftgenerator.concept.nodes.root_distributions._base_distribution import _BaseDistribution

from sid2re.driftgenerator.utils.type_aliases import NumberArray


class ConstantDistribution(_BaseDistribution):
    """Value generating distribution used in root features. This distribution only generates one constant value."""

    @classmethod
    def _generate(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:  WPS123
        """
        Generate a constant value.

        Parameters
        ----------
        w : NumberArray
            Behaviour defining vector of values between 0 and 1.
            w[0]: value in [minimum, maximum] that used as constant function value.
        timestamp : float
            Point in time for which the function value of the function is asked (not used in this case).
        minimum : float
            Minimal function value that is allowed to be generated.
        maximum : float
            Maximal function value that is allowed to be generated.

        Returns
        -------
        float
            The constant value sampled from the uniform distribution defined by w.
        """
        generated_value = (w[0] * (maximum - minimum)) + minimum
        return float(generated_value)



import math

import numpy as np

from sid2re.driftgenerator.concept.nodes.root_distributions._base_distribution import _BaseDistribution
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class PeriodicalDistribution(_BaseDistribution):
    """Value generating distribution used in root features.

    This distributions generation follows a time dependent cycle.
    """

    @classmethod
    def _generate(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:  WPS123
        """
        Generate a value based on a time-dependent cycle.

        Parameters
        ----------
        w : NumberArray
            Behaviour defining vector of values between 0 and 1.
            w[0]: share of the interval used for shifting the function along the time axis.
            w[1]: defining periodical interval between 5 and 50.
            w[2]: share of the [feature_min, feature_max] interval, used for value axis shift.
            w[3]: proportional to pitch of the linear function.
            w[4]: used to determine whether the pitch of the linear function is positive or negative.
        timestamp : float
            Point in time for which the function value of the transformed linear function is asked.
        minimum : float
            Minimal function value that is allowed to be generated.
        maximum : float
            Maximal function value that is allowed to be generated.

        Returns
        -------
        float
            The generated value.
        """
        selection = math.floor(3 * w[0])
        generated_value = 0.0
        if selection <= 0:
            generated_value = cls._dist_cos(w[1:], timestamp, minimum, maximum)
        if selection == 1:
            generated_value = cls._dist_linear(w[1:], timestamp, minimum, maximum)
        if selection >= 2:
            generated_value = cls._dist_quadratic(w[1:], timestamp, minimum, maximum)
        return float(generated_value)

    @classmethod
    def _dist_cos(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:  WPS123
        """
        Generate a value based on a cosine function.

        Parameters
        ----------
        w : NumberArray
            Behaviour defining vector of values between 0 and 1.
            w[0]: share of the interval used for shifting the cosine along the time axis.
            w[1]: defining periodical interval between 5 and 50.
            w[2]: share of the [feature_min, feature_max] interval, used for value axis shift.
            w[3]: proportional to cosine amplitudes.
        timestamp : float
            Point in time for which the function value of the transformed cosine is asked.
        minimum : float
            Minimal function value that is allowed to be generated.
        maximum : float
            Maximal function value that is allowed to be generated.

        Returns
        -------
        float
            The generated value.
        """
        interval = 45 * w[1] + 5
        shift = w[0] * interval
        average = w[2] * (maximum - minimum) + minimum
        amplitude = min(maximum - average, average - minimum) * w[3]

        timestamp = (timestamp + shift) * math.pi
        timestamp = timestamp % interval
        return float((np.cos(timestamp) * amplitude) + average)

    @classmethod
    def _dist_linear(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:  WPS123
        """
        Generate a value based on a linear function.

        Parameters
        ----------
        w : NumberArray
            Behaviour defining vector of values between 0 and 1.
            w[0]: share of the interval used for shifting the function along the time axis.
            w[1]: defining periodical interval between 5 and 50.
            w[2]: share of the [feature_min, feature_max] interval, used for value axis shift.
            w[3]: proportional to pitch of the linear function.
            w[4]: used to determine whether the pitch of the linear function is positive or negative.
        timestamp : float
            Point in time for which the function value of the transformed linear function is asked.
        minimum : float
            Minimal function value that is allowed to be generated.
        maximum : float
            Maximal function value that is allowed to be generated.

        Returns
        -------
        float
            The generated value.
        """
        sign = (math.floor(w[4] - 0.5) * 2) + 1
        interval = 45 * w[1] + 5
        shift = w[0] * interval
        base = w[2] * (maximum - minimum) + minimum
        amplitude = (maximum - base) / interval * w[3]

        timestamp = (timestamp + shift) % interval
        if sign == -1:
            base = base + (maximum - base)
        return float(timestamp * sign * amplitude + base)

    @classmethod
    def _dist_quadratic(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:WPS123
        """
        Generate a value based on a quadratic function.

        Parameters
        ----------
        w : NumberArray
            Behaviour defining vector of values between 0 and 1.
            w[0]: share of the interval used for shifting the function along the time axis.
            w[1]: defining periodical interval between 5 and 50.
            w[2]: share of the [feature_min, feature_max] interval, used for value axis shift.
            w[3]: proportional to pitch of the quadratic function.
            w[4]: used to determine whether the pitch of the quadratic function is positive or negative.
        timestamp : float
            Point in time for which the function value of the transformed quadratic function is asked.
        minimum : float
            Minimal function value that is allowed to be generated.
        maximum : float
            Maximal function value that is allowed to be generated.

        Returns
        -------
        float
            The generated value.
        """
        sign = (math.floor(w[4] - 0.5) * 2) + 1
        interval = 45 * w[1] + 5
        shift = w[0] * interval
        base = w[2] * (maximum - minimum) + minimum
        amplitude = (maximum - base) / (interval ** 2) * w[3]

        timestamp = (timestamp + shift) % interval
        if sign == -1:
            base = base + (maximum - base)
        return float((timestamp ** 2) * sign * amplitude + base)



"""Type aliases and definitions."""
from typing import Any, Callable, Dict, Optional, Tuple, Union

import numpy as np
from numpy import typing as npt

# Base Types
NumberArray = npt.NDArray[np.number]
DataTuple = Tuple[NumberArray, NumberArray]
BoolArray = npt.NDArray[np.bool_]
IntArray = npt.NDArray[np.int_]
Estimator = Any

# Type Compounds.
EnsDict = Dict[int, Estimator]
MetrOut = Union[float, npt.NDArray[np.float64]]
Metric = Callable[[NumberArray, NumberArray, Optional[bool]], MetrOut]



from sid2re.driftgenerator.concept.drift_behaviours._base_drift_behaviour import _BaseDriftBehaviour
from sid2re.driftgenerator.concept.drift_transition_functions import _BaseTransitionFunction
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class FaultySensorDriftBehaviour(_BaseDriftBehaviour):
    """Specify the transformation of the concept coefficients caused by a faulty sensor.

    A faulty sensor shuts down any information gained by the sensor. This is expressed by setting the coefficient to
    zero so that information is deleted.
    The transformation is therefore the inverse of the current coefficients.
    """

    def _compute_coeff_delta(
        self,
        coefficient: NumberArray,
        transition_behaviour: _BaseTransitionFunction,
        current_time: float
    ) -> NumberArray:
        """
        Compute the change in coefficients during drift caused by a faulty sensor.

        Parameters
        ----------
        coefficient : NumberArray
            The coefficients to be modified.
        transition_behaviour : _BaseTransitionFunction
            The behavior of transition.
        current_time : float
            The current time in the system.

        Returns
        -------
        NumberArray
            The computed change in coefficients.
        """
        return -coefficient

    @property
    def _drift_name(self) -> str:
        """
        The name of the drift behavior.

        Returns
        -------
        str
            The name of the drift behavior: "faulty_sensor".
        """
        return "faulty_sensor"



from abc import ABC, abstractmethod


class _BaseTransitionFunction(ABC):

    @classmethod
    def transition_coefficient(cls, current_time: float, shift_centre: float, shift_radius: float) -> float:
        """Compute the transition coefficient.

        Parameters
        ----------
        current_time : float
            The current time in the system.
        shift_centre : float
            The center of the shift.
        shift_radius : float
            The radius around the shift center.

        Returns
        -------
        float
            The transition coefficient, ranging from 0 to 1.

        Raises
        ------
        ValueError
            If shift_radius is negative or zero.

        """
        if current_time < shift_centre - shift_radius:
            return 0
        if current_time > shift_centre + shift_radius:
            return 1
        if shift_radius <= 0:
            raise ValueError("Shift Radius cannot be negative or zero")
        coefficient = cls._transition_coefficient(current_time, shift_centre, shift_radius)
        return coefficient

    @classmethod
    @abstractmethod
    def _transition_coefficient(cls, current_time: float, shift_centre: float, shift_radius: float) -> float:
        """Abstract method to compute the transition coefficient.

        Parameters
        ----------
        current_time : float
            The current time in the system.
        shift_centre : float
            The center of the shift.
        shift_radius : float
            The radius around the shift center.

        Returns
        -------
        float
            The computed transition coefficient.
        """



from typing import List

import numpy as np

from sid2re.driftgenerator.concept.utils import get_random_model
from sid2re.driftgenerator.concept.nodes._base_node import _BaseNode
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class RandomConceptFunctionNode(_BaseNode):
    """Node representing the concept function that uses a subset of all available features to produce target values."""
    def __init__(self, number_of_dependency_models: int, num_dependencies: int, n_outputs: int):
        """
        Initializes a RandomConceptFunctionNode instance.

        Parameters
        ----------
        number_of_dependency_models : int
            Number of models to generate.
        num_dependencies : int
            Number of dependencies.
        n_outputs : int
            Number of output dimensions.
        """

        super().__init__()
        self.config = (np.random.rand(number_of_dependency_models) - 0.5) * 1000 / number_of_dependency_models
        self.collection = []
        self.tag_collection = []
        self.n_outputs = n_outputs

        for _ in range(number_of_dependency_models):
            model_tag, model = get_random_model(num_dependencies, n_outputs)
            self.collection.append(model)
            self.tag_collection.append(model_tag)

    def generate_data(self, time_stamps: NumberArray, inputs: List[NumberArray]) -> NumberArray:
        """
        Used to label inputs dependent on the time with this concept.
        As the concept might change over time, the time_stamps to the inputs have to be provided.

        Parameters
        ----------
        time_stamps : NumberArray
            Points in time of the given inputs. Shape: (n_samples)
        inputs : List[NumberArray]
            Inputs that are to be mapped to the targets. Shape: (n_samples, feature_dim)


        Returns
        -------
        NumberArray
            Target labels to the given inputs. Shape: (n_samples, target_dim)
        """

        init_flag = False
        labels: NumberArray = np.array([])
        input_values = np.dstack(inputs)[0]
        for (sample, stamp) in zip(input_values, time_stamps):
            output = np.zeros((1, self.n_outputs))
            coeffs = self.current_config(stamp)
            for (model, coeffizient, tag) in zip(self.collection, coeffs, self.tag_collection):
                if (tag in ["MLP", "Decision Tree"]) and (len(model.predict(sample.reshape(1, -1)).shape) == 1):
                    model_prediction = model.predict(sample.reshape(1, -1))[:, None]
                else:
                    model_prediction = model.predict(sample.reshape(1, -1))
                output += coeffizient * model_prediction
            if init_flag:
                labels = np.concatenate((labels, output), axis=0)
            else:
                # print(output)
                labels = output
                # print(labels)
                init_flag = True
        if labels.shape[1] == 1:
            labels = labels.flatten()
        return labels



import random
from typing import Any, Tuple
from warnings import simplefilter

import numpy as np
from sklearn.exceptions import ConvergenceWarning
from sklearn.kernel_ridge import KernelRidge
from sklearn.neural_network import MLPRegressor
from sklearn.tree import DecisionTreeRegressor

simplefilter("ignore", category=ConvergenceWarning)


def get_random_model(number_of_features: int, number_of_outputs: int) -> Tuple[str, Any]:
    """Generates a random regression model.

    Parameters
    ----------
    number_of_features : int
        The number of input features.
    number_of_outputs : int
        The number of output targets.

    Returns
    -------
    Tuple[str, Any]
        A tuple containing the model label and the generated regression model.
    """

    selection = random.randint(1, 3)
    if selection == 1:
        # Multilayer Perceptron
        random_state = random.randint(1, 1000)
        features = (np.random.rand(20, number_of_features) - 0.5) * 10
        targets = (np.random.rand(20, number_of_outputs) - 0.5) * 10
        if number_of_outputs == 1:
            targets = targets.flatten()
        return "MLP", MLPRegressor(random_state=random_state, hidden_layer_sizes=random.randint(1, 100)).fit(
            features,
            targets
        )

    if selection == 2:
        # Kernel Regression
        features = (np.random.rand(20, number_of_features) - 0.5) * 10
        targets = (np.random.rand(20, number_of_outputs) - 0.5) * 10
        kernel_ridge = KernelRidge(
            kernel=random.choice(("poly", "polynomial", "rbf", "laplacian", "sigmoid", "cosine"))
        ).fit(features, targets)
        return f"KernelRidge / {kernel_ridge.get_params()['kernel']}", kernel_ridge

    # Standard Output: Tree Regression
    random_state = random.randint(1, 1000)
    features = (np.random.rand(20, number_of_features) - 0.5) * 10
    targets = (np.random.rand(20, number_of_outputs) - 0.5) * 10
    return "Decision Tree", DecisionTreeRegressor(random_state=random_state).fit(features, targets)



from abc import ABC, abstractmethod
from typing import List

from sid2re.driftgenerator.concept.drift_transition_functions import _BaseTransitionFunction
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class _BaseDriftBehaviour(ABC):
    """Abstract base class defining the behavior of drift in a system."""

    def __init__(
        self,
        drift_time: float,
        drift_radius: float,
        coefficient_shift: NumberArray,
        reoccurring: bool
    ) -> None:
        """
        Initialize the drift behavior.

        Parameters
        ----------
        drift_time : float
            The time at which the drift occurs.
        drift_radius : float
            The radius around the drift time within which the drift effect is active.
        coefficient_shift : NumberArray
            The amount by which coefficients are shifted during the drift.
        reoccurring : bool
            Indicates if the drift occurs repeatedly.
        """
        self.drift_time = drift_time
        self.drift_radius = drift_radius
        self.coefficient_shift = coefficient_shift
        self.drift_begin = drift_time - drift_radius
        self.drift_end = drift_time + drift_radius
        self.reoccurring = reoccurring
        if self.reoccurring:
            self.to_drift = self.__class__(drift_time - (drift_radius / 2), drift_radius / 2, coefficient_shift, False)
            self.from_drift = self.__class__(
                drift_time + (drift_radius / 2), drift_radius / 2, -coefficient_shift,
                False
            )

    @property
    def drift_information(self) -> List:
        """Returns information about the drift as a list [Time, Radius, Coefficient,Name]"""
        return [self.drift_time, self.drift_radius, self.coefficient_shift, self._drift_name]

    def modify_coeff(
        self,
        coefficient: NumberArray,
        transition_behaviour: _BaseTransitionFunction,
        current_time: float
    ) -> NumberArray:
        """
        Modify the coefficients based on the drift behavior.

        Parameters
        ----------
        coefficient : NumberArray
            The coefficients to be modified.
        transition_behaviour : _BaseTransitionFunction
            The behavior of transition.
        current_time : float
            The current time in the system.

        Returns
        -------
        NumberArray
            The modified coefficients.
        """
        coefficient = coefficient.copy()  # prevent manipulation of the initial input coefficient
        if self.reoccurring:
            coefficient = self.to_drift.modify_coeff(coefficient, transition_behaviour, current_time)
            coefficient = self.from_drift.modify_coeff(coefficient, transition_behaviour, current_time)
        else:
            if self.drift_begin < current_time <= self.drift_end:
                coefficient = coefficient + self._compute_coeff_delta(coefficient, transition_behaviour, current_time)
            elif current_time > self.drift_end:
                coefficient = coefficient + self.coefficient_shift

        return coefficient

    @abstractmethod
    def _compute_coeff_delta(  # noqa: DOC101,DOC109,DOC103,DOC201,DOC203
        self,
        coefficient: NumberArray,
        transition_behaviour: _BaseTransitionFunction,
        current_time: float
    ) -> NumberArray:
        """
        Abstract method to compute the change in coefficients during drift.

        Parameters:
        -----------
        coefficient : NumberArray
            The coefficients to be updated.
        transition_behaviour : _BaseTransitionFunction
            The transition behavior.
        current_time : float
            The current time in the simulation.

        Returns:
        --------
        NumberArray
            The updated coefficients.
        """

    @property
    @abstractmethod
    def _drift_name(self) -> str:
        """
        Abstract property representing the name of the drift behavior.

        Returns:
        --------
        str
            The name of the drift behavior.
        """



from abc import ABC, abstractmethod

from sid2re.driftgenerator.utils.type_aliases import NumberArray


class _BaseDistribution(ABC):
    """Abstract base class defining the distribution behavior."""

    @classmethod
    def generate(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:  WPS123
        """Generate a value based on the distribution behavior.

        Parameters
        ----------
        w : NumberArray
            The weights for distribution generation.
        timestamp : float
            The timestamp for distribution generation.
        minimum : float
            The minimum value for distribution generation.
        maximum : float
            The maximum value for distribution generation.

        Returns
        -------
        float
            The generated value.
        """
        if not w.any():
            return 0.0
        output = cls._generate(w, timestamp, minimum, maximum)
        assert isinstance(output, float)
        return output

    @classmethod
    @abstractmethod
    def _generate(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:WPS123
        """Abstract method to generate a value based on the distribution behavior.

        Parameters
        ----------
        w : NumberArray
            The weights for distribution generation.
        timestamp : float
            The timestamp for distribution generation.
        minimum : float
            The minimum value for distribution generation.
        maximum : float
            The maximum value for distribution generation.

        Returns
        -------
        float
            The generated value.
        """



from sid2re.driftgenerator.generator import DataGeneratorGraph
from sid2re.driftgenerator.concept.drift_behaviours import SuddenDriftBehaviour, IncrementalDriftBehaviour, \
    GradualDriftBehaviour, FaultySensorDriftBehaviour
import numpy as np
import pytest
import networkx as nx


def test_standard_run():
    generator = DataGeneratorGraph(number_of_models=5,
                                   number_of_dependency_models=2,
                                   n_target_dep=2,
                                   number_of_features=np.array([4, 3]),
                                   number_of_outputs=2,
                                   concept_drifts=1,
                                   data_drifts=1,
                                   number_of_data_points=1000,
                                   noise_var=0.1,
                                   continuous_time=True,
                                   rand_seed=1,
                                   drift_blocking_mode=True,
                                   max_severity=1,
                                   min_severity=1,
                                   root_distros=[1, 1, 1, 1])
    data_df = generator.get_data()
    print(data_df)
    concept_shift_information, data_shift_information = generator.get_shift_information()
    concept = generator.get_concept()
    _ = generator.get_concept_adjacency_matrix()
    _ = generator.get_concept_adjacency_matrix(output_node_names=True)
    print(concept_shift_information)
    print(data_shift_information)
    assert data_df.size == 1000 * 10


@pytest.mark.parametrize("n_cd_drifts", [0, 1, 2])
@pytest.mark.parametrize("n_dd_drifts", [0, 1, 2])
@pytest.mark.parametrize("cd_class", [SuddenDriftBehaviour, IncrementalDriftBehaviour])
@pytest.mark.parametrize("dd_class", [GradualDriftBehaviour, FaultySensorDriftBehaviour])
@pytest.mark.parametrize("bool_switch", [False, True])
def test_parameter_grid(n_cd_drifts, n_dd_drifts, cd_class, dd_class, bool_switch):
    generator = DataGeneratorGraph(number_of_models=5,
                                   number_of_dependency_models=2,
                                   n_target_dep=2,
                                   number_of_features=np.array([4, 3]),
                                   number_of_outputs=2,
                                   concept_drifts=n_cd_drifts,
                                   data_drifts=n_dd_drifts,
                                   concept_drift_class=cd_class,
                                   data_drift_class=dd_class,
                                   number_of_data_points=1000,
                                   noise_var=0.1,
                                   continuous_time=bool_switch,
                                   rand_seed=1,
                                   drift_blocking_mode=bool_switch,
                                   max_severity=1,
                                   min_severity=1,
                                   root_distros=[0, 1, 1, 1],
                                   level_limited=bool_switch,
                                   limit_target_dep=bool_switch,
                                   min_number_dependencies=1,
                                   max_number_dependencies=10
                                   )
    data_df = generator.get_data()
    concept_shift_information, data_shift_information = generator.get_shift_information()
    assert data_df.size == 1000 * 10


def test_false_initialization():
    with pytest.raises(ValueError) as excinfo:
        _ = DataGeneratorGraph(number_of_models=5,
                               number_of_dependency_models=2,
                               n_target_dep=2,
                               number_of_features=np.array([4, 3]),
                               number_of_outputs=2,
                               concept_drifts=1,
                               data_drifts=1,
                               concept_drift_class=SuddenDriftBehaviour,
                               data_drift_class=IncrementalDriftBehaviour,
                               number_of_data_points=1000,
                               noise_var=0.1,
                               continuous_time=True,
                               rand_seed=1,
                               drift_blocking_mode=True,
                               max_severity=1,
                               min_severity=1,
                               root_distros=[4, 1, 1, 1])
    assert len(str(excinfo.value)) > 0


def test_graph_initialization():
    graph = nx.DiGraph()
    graph.add_node("feat_1")
    graph.add_node("feat_2")
    graph.add_node("feat_3")
    graph.add_node("feat_4")
    graph.add_edge("feat_1", "feat_3")
    graph.add_edge("feat_2", "feat_3")
    graph.add_node("label_1")
    graph.add_edge("feat_1", "label_1")
    graph.add_edge("feat_3", "label_1")
    graph.add_edge("feat_4", "label_1")
    generator = DataGeneratorGraph(graph=graph,
                                   output_nodes=["label_1"],
                                   number_of_models=5,
                                   number_of_dependency_models=2,
                                   number_of_outputs=2,
                                   concept_drifts=1,
                                   data_drifts=1,
                                   number_of_data_points=1000,
                                   noise_var=0.1,
                                   continuous_time=True,
                                   rand_seed=1,
                                   drift_blocking_mode=True,
                                   max_severity=1,
                                   min_severity=1,
                                   root_distros=[1, 1, 0, 0])
    data_df = generator.get_data()
    print(data_df)
    concept_shift_information, data_shift_information = generator.get_shift_information()
    assert data_df.size == 1000 * 6



"""Standard and specific error metrics for applications on non-stationary environments."""
from typing import Optional

import numpy as np
from numpy import typing as npt
from sid2re.driftgenerator.utils.type_aliases import Metric, MetrOut, NumberArray


def error(
    y_pred: NumberArray,
    y_truth: NumberArray,
    averaging: Optional[bool] = False,
) -> NumberArray:
    """Calculate the elementwise error for a given prediction and the ground truth labels.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    NumberArray
        Of shape (n_samples,n_labels) or (n_samples)

    Raises
    ------
    ValueError
        If wrong input types are passed.

    """
    if not (isinstance(y_pred, (np.ndarray, np.generic)) and isinstance(y_truth, (np.ndarray, np.generic))):
        raise ValueError(
            f'Tried to compute metric for input instances {type(y_pred)} and {type(y_truth)}.'
            + ' Please only use np.ndarray.',
        )
    if averaging:
        return np.asarray(np.mean(np.subtract(y_truth, y_pred), axis=1), dtype=np.float64)
    return np.subtract(y_truth, y_pred)


def ae(
    y_pred: NumberArray,
    y_truth: NumberArray,
    averaging: Optional[bool] = False,
) -> npt.NDArray[np.float64]:
    """Calculate the elementwise absolute error for a given prediction and the ground truth labels.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    npt.NDArray[np.float64]
        Of shape (n_samples,n_labels) or (n_samples)
    """
    if averaging:
        return np.asarray(np.mean(np.abs(error(y_pred, y_truth)), axis=1), dtype=np.float64)
    return np.asarray(np.abs(error(y_pred, y_truth)), dtype=np.float64)


def mae(
    y_pred: NumberArray,
    y_truth: NumberArray,
    averaging: Optional[bool] = False,
) -> MetrOut:
    """Calculate the mean absolute error for a given prediction and the ground truth labels.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    MetrOut
        Of shape (n_labels) or float
    """
    if averaging:
        return float(np.mean(ae(y_pred, y_truth)))
    return np.asarray(np.mean(ae(y_pred, y_truth), axis=0), dtype=np.float64)


def mse(
    y_pred: NumberArray,
    y_truth: NumberArray,
    averaging: Optional[bool] = False,
) -> MetrOut:
    """Calculate the mean squared error for a given prediction and the ground truth labels.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    MetrOut
        Of shape (n_labels) or float
    """
    if averaging:
        return float(np.mean(np.square(error(y_pred, y_truth))))
    return np.asarray(np.mean(np.square(error(y_pred, y_truth)), axis=0), dtype=np.float64)


def rmse(
    y_pred: NumberArray,
    y_truth: NumberArray,
    averaging: Optional[bool] = False,
) -> MetrOut:
    """Calculate the root mean squared error for a given prediction and the ground truth labels.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    MetrOut
        Of shape (n_labels) or float
    """
    if averaging:
        return float(np.mean(np.sqrt(mse(y_pred, y_truth))))
    return np.asarray(np.sqrt(mse(y_pred, y_truth)), dtype=np.float64)


def ape(
    y_pred: NumberArray,
    y_truth: NumberArray,
    averaging: Optional[bool] = False,
) -> npt.NDArray[np.float64]:
    """Calculate the absolute percentage error for a given prediction and the ground truth labels.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    npt.NDArray[np.float64]
        Of shape (n_samples,n_labels) or (n_samples)
    """
    epsilon = np.finfo(np.float64).eps
    if averaging:
        return np.asarray(np.mean((ae(y_pred, y_truth) / np.maximum(np.abs(y_truth), epsilon)), axis=1))
    return np.asarray(ae(y_pred, y_truth) / np.maximum(np.abs(y_truth), epsilon))


def mape(
    y_pred: NumberArray,
    y_truth: NumberArray,
    averaging: Optional[bool] = False,
) -> MetrOut:
    """Calculate the mean absolute percentage error for a given prediction and the ground truth labels.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    MetrOut
        Of shape (n_labels) or float
    """
    if averaging:
        return float(np.mean(ape(y_pred, y_truth)))
    return np.asarray(np.mean(ape(y_pred, y_truth), axis=0), dtype=np.float64)


def bias(
    y_pred: NumberArray,
    y_truth: NumberArray,
    averaging: Optional[bool] = False,
) -> MetrOut:
    """Calculate the bias error, as the sum of all errors, for a given prediction and the ground truth labels.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    MetrOut
        Of shape (n_labels) or float
    """
    if averaging:
        return float(np.mean(np.sum(error(y_pred, y_truth), axis=0)))
    return np.asarray(np.sum(error(y_pred, y_truth), axis=0), dtype=np.float64)


def drift_correction_error(
    y_pred: NumberArray,
    y_truth: NumberArray,
    local_error: Metric = ae,
    averaging: Optional[bool] = False,
) -> MetrOut:
    """Calculate the error for drift correction methods, as the sum of prediction errors-min(prediction error).

    As this metric tries to evaluate the effectiveness of the drift adaption the best case performance, that is mostly
    dependent on the used prediction model, we only evaluate how much we deviate from the best case/stationary concept
    prediction.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels).
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    local_error : Metric
        Internal metric on which basis to calculate the prediction error and also minimal achieved prediction error.
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    MetrOut
        Of shape (n_labels) or float
    """
    e_t = local_error(y_pred, y_truth, False)

    if averaging:
        return float(np.mean(np.sum(e_t - np.amin(e_t, axis=0), axis=0)))
    return np.asarray(np.sum(e_t - np.amin(e_t, axis=0), axis=0), dtype=np.float64)


def drift_reference_error(
    y_pred: NumberArray,
    y_truth: NumberArray,
    y_ref: NumberArray,
    local_error: Metric = ae,
    averaging: Optional[bool] = False,
) -> MetrOut:
    """Calculate the cumulative difference in performance of a drift adaption model and a reference model.

    If this error is negative the drift adaption model has better performance on average.

    Parameters
    ----------
    y_pred : NumberArray
        Of shape (n_samples,n_labels). Predictions of the base model.
    y_truth : NumberArray
        Of shape (n_samples,n_labels).
    y_ref : NumberArray
        Of shape (n_samples,n_labels). Prediction of the reference model.
    local_error : Metric
        Internal metric on which basis to calculate the prediction error and also minimal achieved prediction error.
    averaging : Optional[bool]
        Whether to average over the label dimensions.

    Returns
    -------
    MetrOut
        Of shape (n_labels) or float
    """
    if averaging:
        return float(
            np.mean(
                np.sum(np.subtract(local_error(y_pred, y_truth, False), local_error(y_ref, y_truth, False)), axis=0),
            ),
        )
    return np.asarray(
        np.sum(np.subtract(local_error(y_pred, y_truth, False), local_error(y_ref, y_truth, False)), axis=0),
        dtype=np.float64,
    )



from sid2re.driftgenerator.concept.nodes.root_distributions.uniform_distribution import UniformDistribution
from sid2re.driftgenerator.concept.nodes.root_distributions.periodical_distribution import PeriodicalDistribution
from sid2re.driftgenerator.concept.nodes.root_distributions.constant_distribution import ConstantDistribution
from sid2re.driftgenerator.concept.nodes.root_distributions.gaussian_distribution import GaussianDistribution
from sid2re.driftgenerator.concept.nodes.root_distributions._base_distribution import _BaseDistribution

__all__ = ["UniformDistribution", "PeriodicalDistribution", "ConstantDistribution", "GaussianDistribution",
           "_BaseDistribution"]



from sid2re.driftgenerator.concept.drift_transition_functions import LinearTransitionFunction
import pytest


@pytest.mark.parametrize("transition_function", [LinearTransitionFunction])
def test_transition_function_compliance(transition_function):
    value = transition_function.transition_coefficient(current_time=0, shift_centre=10, shift_radius=5)
    assert value == 0
    value = transition_function.transition_coefficient(current_time=20, shift_centre=10, shift_radius=5)
    assert value == 1
    value = transition_function.transition_coefficient(current_time=10, shift_centre=10, shift_radius=5)
    assert 0 < value < 1
    with pytest.raises(ValueError) as excinfo:
        _ = transition_function.transition_coefficient(current_time=10, shift_centre=10, shift_radius=0)
    assert len(str(excinfo.value)) > 0


def test_linear_transition_function():
    transition_function = LinearTransitionFunction()
    value = transition_function.transition_coefficient(current_time=2, shift_centre=5, shift_radius=5)
    assert value == 0.2
    value = transition_function.transition_coefficient(current_time=6, shift_centre=5, shift_radius=5)
    assert value == 0.6



import random
from typing import Any, Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd

from sid2re.driftgenerator.concept.drift_behaviours import _BaseDriftBehaviour
from sid2re.driftgenerator.concept.drift_transition_functions import _BaseTransitionFunction
from sid2re.driftgenerator.concept.nodes import RandomConceptFunctionNode, RandomFunctionNode, RootFeatureNode
from sid2re.driftgenerator.concept.nodes.root_distributions import _BaseDistribution
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class ConceptGraph:
    """Build the networkx graph used to express the concept dependencies.

    This graph includes information of depth of dependencies, what are root nodes and what are targets.
    This is the base of truth for the data generation process.
    """

    def __init__(
        self,
        number_of_features: NumberArray,
        number_of_outputs: int,
        number_of_models: int,
        number_of_dependency_models: int,
        min_number_dependencies: int,
        max_number_dependencies: int,
        level_limited: bool,
        limit_target_dep: bool,
        n_target_dep: int,
        feature_min: float = -100.0,
        feature_max: float = 100.0
    ):
        """
        Initializes a ConceptGraph instance.

        Parameters
        ----------
        number_of_features : NumberArray
            Array indicating the number of features at each level.

        number_of_outputs : int
            Number of output nodes.

        number_of_models : int
            Number of models.

        number_of_dependency_models : int
            Number of dependency models.

        min_number_dependencies : int
            Minimum number of dependencies for each node.

        max_number_dependencies : int
            Maximum number of dependencies for each node.

        level_limited : bool
            Whether the dependencies are limited by the level.

        limit_target_dep : bool
            Whether to limit the target dependencies.

        n_target_dep : int
            Number of target dependencies.

        feature_min : float
            Minimum value (default is -100.0).

        feature_max : float
            Maximum value (default is 100.0).
        """
        self.graph: nx.DiGraph = nx.DiGraph()
        self.number_of_features: np.ndarray = number_of_features
        self.node_pos: Dict[str, Tuple[int, int]] = {}
        self.r_map: Dict[str, RootFeatureNode] = {}
        self.fu_map: Dict[str, RandomFunctionNode] = {}
        self.c_map: Dict[str, RandomConceptFunctionNode] = {}
        self.f_map: Dict[str, Union[RootFeatureNode, RandomFunctionNode, RandomConceptFunctionNode]] = {}
        self.data_map: Dict[str, np.ndarray] = {}
        self.level_nodes: List[List[str]] = []
        self.feature_idx: List[str] = []
        self.target_nodes: List[str] = []
        current_node: int = 0

        # Add Nodes to the Graph
        for level_idx in range(number_of_features.shape[0]):
            level_nodes = []
            for feature_idx in range(number_of_features[level_idx]):
                self.graph.add_node(f"{level_idx}_{current_node}")
                self.node_pos[f"{level_idx}_{current_node}"] = (feature_idx, level_idx)
                if level_idx == 0:
                    self.r_map[f"{level_idx}_{current_node}"] = RootFeatureNode(feature_min, feature_max)
                    self.f_map[f"{level_idx}_{current_node}"] = self.r_map[f"{level_idx}_{current_node}"]
                level_nodes += [f"{level_idx}_{current_node}"]
                self.feature_idx += [f"{level_idx}_{current_node}"]
                # self.f_map[f"{l}_{current_node}"] = RandomFunctionNode(number_of_dependency_models)
                # Do later with more information
                current_node += 1
            self.level_nodes.append(level_nodes)
        self.graph.add_node("concept")
        self.node_pos["concept"] = (0, number_of_features.shape[0])
        self.c_map["concept"] = RandomConceptFunctionNode(number_of_models, n_target_dep, number_of_outputs)
        self.f_map["concept"] = self.c_map["concept"]
        # Generate dependencies
        for node in list(self.graph.nodes()):
            # List possible dependencies
            if "0_" in node:
                continue
            if node == "concept":
                num_dependencies = n_target_dep
                if limit_target_dep:
                    potential_depend = [potential_parent_node for potential_parent_node in list(self.graph.nodes()) if
                                        f"{number_of_features.shape[0] - 1}_" in potential_parent_node]
                else:
                    for feature_idx in range(number_of_features.shape[0]):
                        potential_depend += [potential_parent_node for potential_parent_node in list(self.graph.nodes())
                                             if f"{feature_idx}_" in potential_parent_node]
            else:
                num_dependencies = random.randint(min_number_dependencies, max_number_dependencies)
                if level_limited:
                    level = node.split("_")[0]
                    potential_depend = [potential_parent_node for potential_parent_node in list(self.graph.nodes()) if
                                        f"{int(level) - 1}_" in potential_parent_node]
                else:
                    level = node.split("_")[0]
                    potential_depend = []
                    for feature_idx in range(int(level)):
                        potential_depend += [potential_parent_node for potential_parent_node in list(self.graph.nodes())
                                             if f"{feature_idx}_" in potential_parent_node]
            num_dependencies = min(num_dependencies, len(potential_depend))
            if node != "concept" and "0_" not in node:
                self.fu_map[node] = RandomFunctionNode(number_of_dependency_models, num_dependencies)
                self.f_map[node] = self.fu_map[node]
            # Select and include dependencies into graph
            dependencies = random.sample(potential_depend, k=num_dependencies)
            for dep in dependencies:
                self.graph.add_edge(dep, node)
        # For visuals, also add Output level_nodes
        for feature_idx in range(number_of_outputs):
            self.graph.add_node(f"O_{feature_idx}")
            self.node_pos[f"O_{feature_idx}"] = (feature_idx, number_of_features.shape[0] + 1)
            self.graph.add_edge("concept", f"O_{feature_idx}")
            self.target_nodes += [f"O_{feature_idx}"]

    def define_root_distros(self, declar: Dict[int, _BaseDistribution]) -> None:
        """
        Defines the root distributions for specified nodes.

        Parameters
        ----------
        declar : Dict[int, _BaseDistribution]
            Dictionary where keys are node indices and values are distributions.
        """
        for idx in declar:
            self.r_map[f"0_{idx}"].set_distro(declar[idx])

    def define_transition_func(self, func: _BaseTransitionFunction) -> None:
        """
        Defines the transition function for nodes in the graph.

        Parameters
        ----------
        func : _BaseTransitionFunction
            Transition function.
        """
        for level in self.level_nodes:
            for node in level:
                self.f_map[node].set_transition(func)

    def generate_level_data(self, time_stamps: NumberArray, level: int) -> None:
        """
        Generates data for nodes at the specified level.

        Parameters
        ----------
        time_stamps : NumberArray
            Array of time stamps.

        level : int
            Level index.
        """
        self.data_map['time_idx'] = time_stamps
        if level != 0:
            for node in self.level_nodes[level]:
                self.data_map[node] = self.fu_map[node].generate_data(
                    time_stamps,
                    inputs=[self.data_map[pred_node] for pred_node in
                            self.graph.predecessors(node)]
                )
        else:
            for node in self.level_nodes[level]:
                self.data_map[node] = self.r_map[node].generate_data(time_stamps)

    def generate_output(self, time_stamps: NumberArray) -> None:
        """
        Generates output data for the concept.

        Parameters
        ----------
        time_stamps : NumberArray
            Array of time stamps.
        """
        predecessors_data = [self.data_map[pred_node] for pred_node in self.graph.predecessors("concept")]
        output = self.c_map["concept"].generate_data(time_stamps, inputs=predecessors_data)
        output = np.array(output)
        if len(output.shape) == 1:
            for node in self.target_nodes:
                self.data_map[node] = output
        else:
            for node_idx, node in enumerate(self.target_nodes):
                self.data_map[node] = output[:, node_idx]

    def pd_data_readout(self) -> pd.DataFrame:
        """
        Converts data to a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            The DataFrame containing the data.
        """
        return pd.DataFrame.from_dict(self.data_map)

    def define_drift(
        self, concept_shift_info: List[_BaseDriftBehaviour],
        data_shift_info: Dict[str, List[_BaseDriftBehaviour]]
    ) -> None:
        """
        Defines drift for nodes in the graph.

        Parameters
        ----------
        concept_shift_info : List[_BaseDriftBehaviour]
            Concept shift information.

        data_shift_info : Dict[str, List[_BaseDriftBehaviour]]
            Data shift information.
        """
        for level in self.level_nodes:
            for node in level:
                self.f_map[node].set_drift(data_shift_info[node])
        self.f_map["concept"].set_drift(concept_shift_info)

    def get_graph(self) -> nx.DiGraph:
        """
        Gets the underlying graph.

        Returns
        -------
        nx.DiGraph
            Directed graph representing the concept.
        """
        return self.graph

    def plot_graph(self) -> None:
        """
        Plots the graph for visualization.
        """
        plt.subplot(111)
        nx.draw(self.graph, pos=self.node_pos, with_labels=True, font_weight='bold')
        plt.show()


def _analyze_graph(graph: nx.DiGraph, output_nodes: List) -> Tuple[NumberArray, List]:
    roots = [node for node in graph.nodes() if graph.in_degree(node) == 0]
    graph.remove_nodes_from(output_nodes)
    last_level = roots
    levels = [roots]
    while len(graph.nodes) > 0:
        graph.remove_nodes_from(last_level)
        current_level = [node for node in graph.nodes() if graph.in_degree(node) == 0]
        levels += [current_level]
        last_level = current_level
    return np.array([len(level) for level in levels]), levels


class ConceptReader(ConceptGraph):
    """ Concept graph representation that is infered from a provided networkx graph.

    This will try to infer information from the provided structure, but might need additional specifications if the
    structure is not unambiguously.
    """

    def __init__(
        self, input_graph: nx.DiGraph, number_of_models: int, number_of_dependency_models: int,
        feature_min: float, feature_max: float, output_nodes: List[Any]
    ):  # pylint: disable=super-init-not-called
        """
        Initializes a ConceptReader instance.

        Parameters
        ----------
        input_graph : nx.DiGraph
            The input graph.
        number_of_models : int
            Number of models.
        number_of_dependency_models : int
            Number of dependency models.
        feature_min : float
            Minimum feature value.
        feature_max : float
            Maximum feature value.
        output_nodes : List[Any]
            List of output nodes.
        """

        self.r_map: Dict[str, RootFeatureNode] = {}
        self.fu_map: Dict[str, RandomFunctionNode] = {}
        self.c_map: Dict[str, RandomConceptFunctionNode] = {}
        self.f_map: Dict[str, Union[RootFeatureNode, RandomFunctionNode, RandomConceptFunctionNode]] = {}
        self.graph = nx.DiGraph()
        number_of_features, hierachy_list = _analyze_graph(input_graph.copy(), output_nodes)
        self.number_of_features = number_of_features
        self.node_pos = {}
        self.old_to_new_map = {}
        self.data_map = {}
        self.level_nodes = []
        self.feature_idx = []
        self.target_nodes = []
        current_node = 0
        # Add Nodes to the Graph
        for level_idx in range(number_of_features.shape[0]):
            level_nodes = []
            for feature_idx, node in zip(range(number_of_features[level_idx]), hierachy_list[level_idx]):
                self.graph.add_node(f"{level_idx}_{current_node}")
                self.old_to_new_map[node] = f"{level_idx}_{current_node}"
                self.node_pos[f"{level_idx}_{current_node}"] = (feature_idx, level_idx)
                if level_idx == 0:
                    self.r_map[f"{level_idx}_{current_node}"] = RootFeatureNode(feature_min, feature_max)
                    self.f_map[f"{level_idx}_{current_node}"] = self.r_map[f"{level_idx}_{current_node}"]
                else:
                    self.fu_map[f"{level_idx}_{current_node}"] = \
                        RandomFunctionNode(number_of_dependency_models, num_dependencies=input_graph.in_degree(node))
                    self.f_map[f"{level_idx}_{current_node}"] = self.fu_map[f"{level_idx}_{current_node}"]
                level_nodes += [f"{level_idx}_{current_node}"]
                self.feature_idx += [f"{level_idx}_{current_node}"]
                current_node += 1
            self.level_nodes.append(level_nodes)
        # Generate dependencies
        for level in hierachy_list:
            for node in level:
                for predec in input_graph.predecessors(node):
                    self.graph.add_edge(self.old_to_new_map[predec], self.old_to_new_map[node])
        concept_inputs = []
        for output in output_nodes:
            for predec in input_graph.predecessors(output):
                concept_inputs += [predec]
        concept_inputs = list(set(concept_inputs))
        self.graph.add_node("concept")
        self.node_pos["concept"] = (0, number_of_features.shape[0])
        self.c_map["concept"] = RandomConceptFunctionNode(number_of_models, len(concept_inputs), len(output_nodes))
        self.f_map["concept"] = self.c_map["concept"]
        for input_values in concept_inputs:
            self.graph.add_edge(self.old_to_new_map[input_values], "concept")
        # For visuals, also add Output level_nodes
        for feature_idx in range(len(output_nodes)):
            self.graph.add_node(f"O_{feature_idx}")
            self.node_pos[f"O_{feature_idx}"] = (feature_idx, number_of_features.shape[0] + 1)
            self.graph.add_edge("concept", f"O_{feature_idx}")
            self.target_nodes += [f"O_{feature_idx}"]
        print("The input graph was remapped to confine to the internal notations. The following maps old to new nodes")
        print(self.old_to_new_map)

    @property
    def number_of_features_per_level(self) -> NumberArray:
        return np.array(self.number_of_features)



from sid2re.driftgenerator.concept.drift_behaviours._base_drift_behaviour import _BaseDriftBehaviour
from sid2re.driftgenerator.concept.drift_behaviours.faulty_sensor_drift_behaviour import FaultySensorDriftBehaviour
from sid2re.driftgenerator.concept.drift_behaviours.gradual_drift_behaviour import GradualDriftBehaviour
from sid2re.driftgenerator.concept.drift_behaviours.incremental_drift_behaviour import IncrementalDriftBehaviour
from sid2re.driftgenerator.concept.drift_behaviours.sudden_drift_behaviour import SuddenDriftBehaviour

__all__ = ["_BaseDriftBehaviour", "FaultySensorDriftBehaviour", "GradualDriftBehaviour", "IncrementalDriftBehaviour",
           "SuddenDriftBehaviour"]



from sid2re.driftgenerator.concept.nodes import RandomConceptFunctionNode, RandomFunctionNode, RootFeatureNode
from sid2re.driftgenerator.concept.nodes.root_distributions import UniformDistribution
from sid2re.driftgenerator.concept.drift_transition_functions import LinearTransitionFunction
from sid2re.driftgenerator.concept.drift_behaviours import SuddenDriftBehaviour, FaultySensorDriftBehaviour, \
    GradualDriftBehaviour, IncrementalDriftBehaviour
import pytest
import numpy as np


@pytest.mark.parametrize("n_models", [5, 10])
@pytest.mark.parametrize("n_dependencies", [1, 2])
@pytest.mark.parametrize("n_outputs", [1, 2])
def test_concept_node_without_drift(n_models, n_dependencies, n_outputs):
    node = RandomConceptFunctionNode(number_of_dependency_models=n_models, num_dependencies=n_dependencies,
                                     n_outputs=n_outputs)
    data = node.generate_data(time_stamps=np.arange(100), inputs=np.random.normal(3, 2.5, size=(n_dependencies, 100)))
    assert data.shape[0] == 100
    if n_outputs != 1:
        assert data.shape[1] == n_outputs


@pytest.mark.parametrize("n_models", [5, 10])
@pytest.mark.parametrize("n_dependencies", [1, 2])
def test_function_node_without_drift(n_models, n_dependencies):
    node = RandomFunctionNode(number_of_dependency_models=n_models, num_dependencies=n_dependencies)
    data = node.generate_data(time_stamps=np.arange(100), inputs=np.random.normal(3, 2.5, size=(n_dependencies, 100)))
    assert data.shape[0] == 100


def test_root_node_without_drift():
    node = RootFeatureNode(minimum=-2, maximum=8)
    node.set_distro(UniformDistribution())
    data = node.generate_data(time_stamps=np.arange(100))
    assert np.amin(data) > -2
    assert np.amax(data) < 8
    assert data.shape[0] == 100


@pytest.mark.parametrize("n_dependencies", [1, 2])
@pytest.mark.parametrize("n_outputs", [1, 2])
@pytest.mark.parametrize("drift_behaviour", [SuddenDriftBehaviour, FaultySensorDriftBehaviour, GradualDriftBehaviour,
                                             IncrementalDriftBehaviour])
@pytest.mark.parametrize("reoccurring", [True, False])
def test_concept_node_with_drift(n_dependencies, n_outputs, drift_behaviour, reoccurring):
    node = RandomConceptFunctionNode(number_of_dependency_models=5, num_dependencies=n_dependencies,
                                     n_outputs=n_outputs)
    drifts = [drift_behaviour(drift_time=50, drift_radius=10, coefficient_shift=np.array([1, 1, 1, 1, 1]),
                              reoccurring=reoccurring)]
    node.set_drift(drifts)
    data = node.generate_data(time_stamps=np.arange(100), inputs=np.random.normal(3, 2.5, size=(n_dependencies, 100)))
    assert data.shape[0] == 100
    if n_outputs != 1:
        assert data.shape[1] == n_outputs


@pytest.mark.parametrize("n_dependencies", [1, 2])
@pytest.mark.parametrize("drift_behaviour", [SuddenDriftBehaviour, FaultySensorDriftBehaviour, GradualDriftBehaviour,
                                             IncrementalDriftBehaviour])
@pytest.mark.parametrize("reoccurring", [True, False])
def test_function_node_with_drift(n_dependencies, drift_behaviour, reoccurring):
    node = RandomFunctionNode(number_of_dependency_models=5, num_dependencies=n_dependencies)
    node.set_transition(LinearTransitionFunction())
    drifts = [
        drift_behaviour(drift_time=50, drift_radius=10, coefficient_shift=np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]),
                        reoccurring=reoccurring)]
    node.set_drift(drifts)
    data = node.generate_data(time_stamps=np.arange(100), inputs=np.random.normal(3, 2.5, size=(n_dependencies, 100)))
    assert data.shape[0] == 100



# Import external functionalities
import random
from typing import Any, Dict, List, Optional, Tuple, Union

import networkx as nx
import numpy as np
import pandas as pd

from sid2re.driftgenerator.concept.drift_behaviours import _BaseDriftBehaviour, FaultySensorDriftBehaviour, \
    GradualDriftBehaviour, IncrementalDriftBehaviour, SuddenDriftBehaviour
from sid2re.driftgenerator.concept.drift_transition_functions import _BaseTransitionFunction, LinearTransitionFunction
from sid2re.driftgenerator.concept.graph_concept import ConceptGraph, ConceptReader
# Import internal functionalities
from sid2re.driftgenerator.concept.nodes.root_distributions import _BaseDistribution, ConstantDistribution, \
    GaussianDistribution, \
    PeriodicalDistribution, UniformDistribution
from sid2re.driftgenerator.utils.type_aliases import NumberArray


def _distro_declarator(
    n_features: int,
    n_uniform_feat: int,
    n_gauss_feat: int,
    n_constant_feat: int,
    n_periodical_feat: int
) -> Dict[int, _BaseDistribution]:
    if n_uniform_feat + n_gauss_feat + n_constant_feat + n_periodical_feat > n_features:
        n_set_feat = n_uniform_feat + n_gauss_feat + n_constant_feat + n_periodical_feat
        raise ValueError(
            f"Too many root features set staticaly: {n_set_feat} of {n_features}"
        )
    declarator: Dict[int, _BaseDistribution] = {}
    indices = np.arange(0, n_features)
    np.random.shuffle(indices)
    list_of_indices = list(indices)

    # set feature behaviour
    for _ in range(n_uniform_feat):
        declarator[list_of_indices.pop()] = UniformDistribution()
    for _ in range(n_gauss_feat):
        declarator[list_of_indices.pop()] = GaussianDistribution()
    for _ in range(n_constant_feat):
        declarator[list_of_indices.pop()] = ConstantDistribution()
    for _ in range(n_periodical_feat):
        declarator[list_of_indices.pop()] = PeriodicalDistribution()

    for _ in range(len(list_of_indices)):
        declarator[list_of_indices.pop()] = random.choice(
            (UniformDistribution(), GaussianDistribution(), ConstantDistribution(), PeriodicalDistribution())
        )

    return declarator


def set_concept_drifts(
    rand_seed: float,
    drift_blocking_mode: bool,
    concept_drifts: int,
    time_stamp: NumberArray,
    min_severity: float,
    max_severity: float,
    number_of_models: int,
    concept_drift_class: Optional[_BaseDriftBehaviour]
) -> List:
    drifts = []
    if drift_blocking_mode and concept_drifts != 0:
        step = (time_stamp[-1] - time_stamp[0]) / concept_drifts
        concept_shift_stamps = np.array([0.5 * step + step * cd_idx for cd_idx in range(concept_drifts)])
    else:
        concept_shift_stamps = np.random.rand(concept_drifts) * (time_stamp[-1] - time_stamp[0])
    if rand_seed != -1:
        random.seed(rand_seed)
        np.random.seed(rand_seed)
    concept_shift_info = []
    drift_spacing = (time_stamp[-1] - time_stamp[0]) / concept_drifts if concept_drifts > 0 else None
    if concept_shift_stamps.size > 0:
        for _ in range(concept_shift_stamps.shape[0]):
            if drift_blocking_mode:
                concept_shift_info.append(
                    [
                        random.random() / 2 * drift_spacing,
                        (np.minimum(
                            max_severity,
                            np.maximum(min_severity, np.random.rand(number_of_models))
                        ) - 0.5) * 1000 / number_of_models,
                        random.choice(
                            [
                                FaultySensorDriftBehaviour, GradualDriftBehaviour,
                                IncrementalDriftBehaviour, SuddenDriftBehaviour
                            ]
                        ),
                        random.choice([True, False]),
                    ]
                )
            else:
                concept_shift_info.append(
                    [
                        random.random() * (np.max(concept_shift_stamps) / 2),
                        (np.random.randint(1) * 2 - 1) * np.minimum(
                            max_severity, np.maximum(min_severity, np.random.rand(number_of_models))
                        ) * 1000 / number_of_models,
                        random.choice(
                            [FaultySensorDriftBehaviour, GradualDriftBehaviour,
                             IncrementalDriftBehaviour, SuddenDriftBehaviour]
                        ),
                        random.choice([True, False])
                    ]
                )
    for stamp, info in zip(concept_shift_stamps, concept_shift_info):
        if concept_drift_class is not None:
            cd_class = concept_drift_class
        else:
            cd_class = info[2]
        drifts += [cd_class(
            drift_time=stamp,
            drift_radius=info[0],
            coefficient_shift=info[1],
            reoccurring=info[3]
        )]  # type: ignore[operator]
    return drifts


def set_data_drifts(
    rand_seed: int,
    drift_blocking_mode: bool,
    data_drifts: int,
    time_stamp: NumberArray,
    feature_idxs: List[str],
    min_severity: float,
    max_severity: float,
    data_drift_class: Optional[_BaseDriftBehaviour]
) -> Dict[str, List[_BaseDriftBehaviour]]:
    drifts: Dict[str, List[_BaseDriftBehaviour]] = {}
    for idx in feature_idxs:
        drifts[idx] = []
    if drift_blocking_mode and data_drifts != 0:
        step = (time_stamp[-1] - time_stamp[0]) / data_drifts
        data_shifts = np.array([0.5 * step + step * dd_idx for dd_idx in range(data_drifts)])
    else:
        data_shifts = np.random.rand(data_drifts) * (time_stamp[-1] - time_stamp[0])
    if rand_seed != -1:
        random.seed(rand_seed)
        np.random.seed(rand_seed)
    data_shift_info = []
    for _ in range(data_drifts):

        if drift_blocking_mode:
            data_shift_info.append(
                [
                    random.choice(feature_idxs),
                    random.random() / 2 * (time_stamp[-1] - time_stamp[0]) / data_drifts,
                    np.minimum(max_severity, np.maximum(min_severity, np.random.rand(10))),
                    random.choice(
                        [
                            FaultySensorDriftBehaviour, GradualDriftBehaviour,
                            IncrementalDriftBehaviour, SuddenDriftBehaviour
                        ]
                    ),
                    random.choice([True, False])
                ]
            )
        else:
            data_shift_info.append(
                [
                    random.choice(feature_idxs),
                    random.random() * (np.max(data_shifts) / 2),
                    np.minimum(max_severity, np.maximum(min_severity, np.random.rand(10))),
                    random.choice(
                        [FaultySensorDriftBehaviour, GradualDriftBehaviour,
                         IncrementalDriftBehaviour, SuddenDriftBehaviour]
                    ),
                    random.choice([True, False])
                ]
            )
    for stamp, info in zip(data_shifts, data_shift_info):
        if data_drift_class is not None:
            dd_class = data_drift_class
        else:
            dd_class = info[3]
        drifts[info[0]] += [
            dd_class(
                drift_time=stamp,
                drift_radius=info[1],
                coefficient_shift=info[2],
                reoccurring=info[4]
            )]  # type: ignore[operator]
    return drifts


class DataGeneratorGraph:
    """
    Generator used to create a non-stationary regression problem.

    Initialization defines the behavior of the non-stationary concept, and sampling defines the behavior of the
    input data.
    """

    def __init__(
        self,
        number_of_data_points: int = 100,
        number_of_features: Optional[NumberArray] = None,
        number_of_outputs: int = 1,
        feature_min: float = -100.0,
        feature_max: float = 100.0,
        number_of_models: int = 5,
        noise_var: float = 0,
        rand_seed: int = -1,
        concept_drifts: int = 0,
        data_drifts: int = 0,
        concept_drift_class: Optional[_BaseDriftBehaviour] = None,
        data_drift_class: Optional[_BaseDriftBehaviour] = None,
        transition_func: Optional[_BaseTransitionFunction] = None,
        continuous_time: bool = True,
        max_time_sparsity: int = 10,
        drift_blocking_mode: bool = False,
        max_severity: float = 1,
        min_severity: float = 0,
        number_of_dependency_models: int = 2,
        min_number_dependencies: int = 2,
        max_number_dependencies: int = 2,
        level_limited: bool = True,
        limit_target_dep: bool = True,
        n_target_dep: int = 1,
        root_distros: Optional[List[int]] = None,
        graph: Optional[nx.DiGraph] = None,
        output_nodes: Optional[List[Any]] = None
    ) -> None:
        """
        Initialize the DataGenerator, defining the behavior of the non-stationary concept.

        Parameters
        ----------
        number_of_data_points : int
            Number of instances included in the dataset generated by `get_data()`.

        number_of_features : Optional[NumberArray]
            Number of input features generated and used in the data stream. Positions in the array define the different
            levels of dependency. The default value only generates two root features that do not have any dependencies.

        number_of_outputs : int
            Number of output targets produced by the artificial non-stationary concept, based on the generated input
            features.

        feature_min : float
            The initial set of generated input features will not fall below this value. Note that data drifts and noise
            may lead to feature values that do not comply with this limitation.

        feature_max : float
            The initial set of generated input features will not surpass this value. Note that data drifts and noise
            may lead to feature values that do not comply with this limitation.

        number_of_models : int
            Number of random models used to build the artificial concept. More models enable more complex concepts.
            Note that an increasing number of models does increase the maximal complexity but also may reduce variance
            in the generated concept, as the random models are linearly combined to form the concept.

        noise_var : float
            Variance of the normal distribution used to sample the scaling factor for noise in features and targets.
            As input and output values may have very different value ranges, we opted for a noisy scaling factor instead
            of an additive noise.

        rand_seed : int
            Seed used for random choices like sampling. If set to -1, no specific seed will be set, leading to random
            results in every execution.

        concept_drifts : int
            Number of concept drifts introduced in the generated dataset of `get_data()`.

        data_drifts : int
            Number of data drifts introduced in the generated dataset of `get_data()`.

        concept_drift_class : Optional[_BaseDriftBehaviour]
            The choosen drift behaviour class (like 'sudden', 'gradual', 'incremental', ...) of the generated
            concept drifts.
            If set to None, the type will be decided for each drift at random.

        data_drift_class : Optional[_BaseDriftBehaviour]
            The choosen drift behaviour class ('sudden', 'gradual', 'incremental', 'faulty_sensor', ...) of the
            generated data drifts.
            If set to None, the type will be decided for each drift at random.

        transition_func : Optional[_BaseTransitionFunction]
            How two concepts are to be blended. Used for incremental and gradual features. Custom transition functions
            may be provided but must implement the API of the following exemplary call:
            `transition_func(current_time_stamp: int, shift_centre: int, shift_radius: float) -> float`.
            Note that the transition function is only called if there is a transition ongoing, i.e.:
            `shift_centre - shift_radius < current_time_stamp <= shift_centre + shift_radius`.

            See the exemplary `LinearTransitionFunction` used as default:
            ```
            def linear_trans_func(current_time, shift_centre, shift_radius):
                return (current_time - (shift_centre - shift_radius)) / (shift_radius * 2)
            ```

        continuous_time : bool
            Whether to use complete time series or time series with missing timestamps.
            If set to False, behavior of missing timestamps can be set by `max_time_sparsity`.

        max_time_sparsity : int
            Only used when `continuous_time=False`.
            Sets the maximal number of timestamps skipped after a timestamp is evaluated.

        drift_blocking_mode : bool
            Whether drifts may overlap or be confined to equally sized timeframes in the generated time series.
            If set to True, concept drifts will not overlap with other concept drifts; same goes for data drifts.
            Note that concept and data drifts are still able to overlap and influence each other.

        max_severity : float
            Internal scaling factor on how drastic drifts are allowed to be.
            Can be used to control the relative intensity of drifts.
            If `max_severity` = `min_severity`, then all drifts will have comparable impacts on data distributions and
            concept.

        min_severity : float
            Internal scaling factor on how drastic drifts have to be.

        number_of_dependency_models : int
            Number of dependency models, used between features of different dependency layers.

        min_number_dependencies : int
            Minimum number of feature dependencies.

        max_number_dependencies : int
            Maximum number of feature dependencies.

        level_limited : bool
            Whether the dependencies are limited to level l to level l-1 or not.

        limit_target_dep : bool
            Whether to limit target dependencies to the last feature layer.

        n_target_dep : int
            Number of target dependencies.

        root_distros : Optional[List[int]]
            List of root distributions. `[n_uniform_feat, n_gauss_feat, n_constant_feat, n_periodical_feat]`
            The input arguments can be used to set the minimal number of features that are sampled from a particular
            kind of distribution. You may only assign as many features as were defined in the first feature layer.
            If there are fewer given assignments than features, the remaining features will be sorted into categories
            by random.

        graph : Optional[nx.DiGraph]
            Concept graph.

        output_nodes : Optional[List[Any]]
            List of output nodes of the input graph.
        """
        if number_of_features is None:
            number_of_features = np.array([2])
        if root_distros is None:
            root_distros = [2, 0, 0, 0]
        if transition_func is None:
            transition_func = LinearTransitionFunction()
        if output_nodes is None:
            output_nodes = []
        # define time index
        if rand_seed != -1:
            random.seed(rand_seed)
            np.random.seed(rand_seed)
        if continuous_time:
            time_stamp = np.arange(0, number_of_data_points)
        else:
            temp_time_stamp = []
            current_time_stamp = 0
            for _ in range(number_of_data_points):
                temp_time_stamp.append(current_time_stamp)
                current_time_stamp += random.randint(1, max_time_sparsity)
            time_stamp = np.array(temp_time_stamp)
        self.continous_time = continuous_time
        self.number_of_data_points = number_of_data_points
        self.time_stamp = time_stamp
        self.transition_func = transition_func
        # define concept graph
        if graph is None:
            self.concept = ConceptGraph(
                number_of_features, number_of_outputs, number_of_models,
                number_of_dependency_models, min_number_dependencies, max_number_dependencies,
                level_limited, limit_target_dep, n_target_dep, feature_min, feature_max
            )
        else:
            self.concept = ConceptReader(
                graph, number_of_models, number_of_dependency_models, feature_min, feature_max,
                output_nodes
            )
            number_of_features = self.concept.number_of_features_per_level
        declar = _distro_declarator(number_of_features[0], *root_distros)
        self.concept.define_root_distros(declar)
        self.concept.define_transition_func(transition_func)
        # define drifts
        self.concept_drifts = set_concept_drifts(
            rand_seed, drift_blocking_mode,
            concept_drifts, time_stamp,
            min_severity, max_severity,
            number_of_models, concept_drift_class
        )
        self.data_drifts = set_data_drifts(
            rand_seed, drift_blocking_mode, data_drifts,
            time_stamp, self.concept.feature_idx,
            min_severity, max_severity, data_drift_class
        )
        self.concept.define_drift(self.concept_drifts, self.data_drifts)
        self.min = feature_min
        self.max = feature_max
        self.n_models = number_of_models
        self.n_features = number_of_features
        self.n_outputs = number_of_outputs
        self.noise_var = noise_var

    def get_data(self) -> pd.DataFrame:
        """
        Use the initialized Generator to sample data points in the non-stationary environment.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame containing the features and targets.

        Notes
        -----
        This method generates a dataset based on the non-stationary environment settings configured during the
        initialization of the `DataGenerator` instance. The data points are sampled according to the defined
        features, models, drifts, and other parameters, ensuring that the resulting dataset reflects the
        specified non-stationary behaviors.
        """

        for feature_idx in range(self.n_features.shape[0]):
            self.concept.generate_level_data(self.time_stamp, feature_idx)
        self.concept.generate_output(self.time_stamp)
        data_df = self.concept.pd_data_readout()
        data_df = data_df.multiply(
            np.concatenate(
                [
                    np.ones((data_df.shape[0], 1)),
                    np.random.normal(1, self.noise_var, (data_df.shape[0], data_df.shape[1] - 1))
                ],
                axis=1
            )
        )
        return data_df

    def get_concept(self) -> ConceptGraph:
        """
        Provides the drift-induced concept function that is used to map the generated inputs to the targets.

        Returns
        -------
        ConceptGraph
            Concept used in the drifted datastream
        """
        return self.concept

    def get_concept_adjacency_matrix(
        self, output_node_names: bool = False,
    ) -> Union[np.ndarray, Tuple[np.ndarray, List[Any]]]:
        """
        Returns the adjacency matrix of the concept graph.

        Parameters
        ----------
        output_node_names : bool
            Whether to include output node names.

        Returns
        -------
        Union[np.ndarray, Tuple[np.ndarray, List[Any]]]
            If `output_node_names` is True, returns a tuple containing:
            - np.ndarray: The adjacency matrix of the concept graph.
            - List[Any]: A list of output node names.

            Otherwise, returns only the adjacency matrix (np.ndarray).

        Notes
        -----
        The adjacency matrix represents the structure of the concept graph where each element (i, j) indicates the
        presence of a directed edge from node i to node j. The `output_node_names` parameter can be used to retrieve
        the names of the nodes in the graph, which can be helpful for interpreting the structure of the graph.
        """
        if output_node_names:
            return np.asarray(nx.adjacency_matrix(self.concept.get_graph()).todense()), self.concept.get_graph().nodes()
        return np.asarray(nx.adjacency_matrix(self.concept.get_graph()).todense())

    def get_shift_information(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Provides two pandas dataframes, that describe the concept and data drifts.
        For the concept drifts (time, duration, weight shift, class) is provided.
        For data drifts (time, affected feature, duration, weight shift, class) is provided.

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            Drift Information DataFrames
        """
        concept_info = []
        for concept_drift in self.concept_drifts:
            concept_info += [concept_drift.drift_information]
        concept = pd.DataFrame(concept_info, columns=['time_stamp(centre)', 'radius', 'shift', 'class'])

        data_info = []
        for node_idx, entry in self.data_drifts.items():
            for data_drift in entry:
                data_info += [[*(data_drift.drift_information), node_idx]]
        features = pd.DataFrame(
            data_info,
            columns=['time_stamp(centre)', 'radius', 'shift of distribution parameters', 'class',
                     'affected_feature']
        )
        return concept, features



from abc import ABC
from typing import List

import numpy as np

from sid2re.driftgenerator.concept.drift_behaviours import _BaseDriftBehaviour
from sid2re.driftgenerator.concept.drift_transition_functions import _BaseTransitionFunction, LinearTransitionFunction
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class _BaseNode(ABC):

    def __init__(self) -> None:
        """
        Initializes a node.
        """

        self.config: NumberArray = np.random.rand(10)
        self.trans: _BaseTransitionFunction = LinearTransitionFunction()
        self.shift_stamps: np.ndarray = np.array([])
        self.shift_info: np.ndarray = np.array([])
        self.drifts: List[_BaseDriftBehaviour] = []

    def set_transition(self, func: _BaseTransitionFunction) -> None:
        """
        Sets the transition function.

        Parameters
        ----------
        func : _BaseTransitionFunction
            The transition function.
        """

        self.trans = func

    def current_config(self, time: float) -> np.ndarray:
        """
        Calculates the current configuration.

        Parameters
        ----------
        time : float
            The current time.

        Returns
        -------
        np.ndarray
            The current configuration.
        """

        coeff = self.config
        for drift in self.drifts:
            coeff = drift.modify_coeff(coefficient=coeff, transition_behaviour=self.trans, current_time=time)
        return coeff

    def set_drift(self, drifts: List[_BaseDriftBehaviour]) -> None:
        """Sets the drift based on the provided shifts, shift_info, and index.

        Parameters
        ----------
        drifts : List[_BaseDriftBehaviour]
            List of drift behaviors.

        Raises
        ------
        TypeError
            If drifts is not a list.
        """
        if not isinstance(drifts, list):
            raise TypeError("Drifts must be a list.")
        self.drifts = drifts



import pytest
from sid2re.driftgenerator.concept.drift_behaviours import FaultySensorDriftBehaviour, GradualDriftBehaviour, \
    IncrementalDriftBehaviour, SuddenDriftBehaviour
from sid2re.driftgenerator.concept.drift_transition_functions import LinearTransitionFunction
import numpy as np

list_of_all_drift_behaviours = [FaultySensorDriftBehaviour, GradualDriftBehaviour,
                                IncrementalDriftBehaviour, SuddenDriftBehaviour]


@pytest.mark.parametrize("behaviour", list_of_all_drift_behaviours)
def test_non_reoccurring_drift_outer_behaviour(behaviour):
    coefficient = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    behaviour = behaviour(50, 10, coefficient_shift=coefficient.copy(), reoccurring=False)
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=10)
    assert (value == coefficient).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=70)
    assert (value == coefficient + 1).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()


def test_non_reoccurring_sudden_drift():
    behaviour = SuddenDriftBehaviour
    coefficient = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    behaviour = behaviour(50, 10, coefficient_shift=coefficient.copy(), reoccurring=False)
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=41)
    assert (value == coefficient + 1).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()


def test_non_reoccurring_faulty_sensor_drift():
    behaviour = FaultySensorDriftBehaviour
    coefficient = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    behaviour = behaviour(50, 10, coefficient_shift=coefficient.copy(), reoccurring=False)
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=41)
    assert (value == coefficient - 1).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=70)
    assert (value == coefficient + 1).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()


def test_non_reoccurring_gradual_drift():
    behaviour = GradualDriftBehaviour
    coefficient = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    behaviour = behaviour(50, 10, coefficient_shift=coefficient.copy(), reoccurring=False)
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=50)
    assert (value == coefficient).all() or (value == coefficient + 1).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()


def test_non_reoccurring_incremental_drift():
    behaviour = IncrementalDriftBehaviour
    coefficient = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    behaviour = behaviour(50, 10, coefficient_shift=coefficient.copy(), reoccurring=False)
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=50)
    assert (value == coefficient + 0.5).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()


@pytest.mark.parametrize("behaviour", list_of_all_drift_behaviours)
def test_reoccurring_drift_outer_behaviour(behaviour):
    coefficient = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    behaviour = behaviour(50, 10, coefficient_shift=coefficient.copy(), reoccurring=True)
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=10)
    assert (value == coefficient).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()
    value = behaviour.modify_coeff(coefficient=coefficient, transition_behaviour=LinearTransitionFunction(),
                                   current_time=70)
    assert (value == coefficient).all()
    assert (coefficient == np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0])).all()



import random

from sid2re.driftgenerator.concept.drift_behaviours._base_drift_behaviour import _BaseDriftBehaviour
from sid2re.driftgenerator.concept.drift_transition_functions import _BaseTransitionFunction
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class GradualDriftBehaviour(_BaseDriftBehaviour):
    """Specify the transformation of the concept coefficients caused by a gradual drift.

    A gradual drift presents a probabilistic transition to a new concept. This is expressed by randomly choosing one
    of the concepts and providing the required coefficients for it.
    """

    def _compute_coeff_delta(
        self,
        coefficient: NumberArray,
        transition_behaviour: _BaseTransitionFunction,
        current_time: float
    ) -> NumberArray:
        """
        Compute the change in coefficients during gradual drift.

        Parameters
        ----------
        coefficient : NumberArray
            The coefficients to be modified.
        transition_behaviour : _BaseTransitionFunction
            The behavior of transition.
        current_time : float
            The current time in the system.

        Returns
        -------
        NumberArray
            The computed change in coefficients.
        """
        prob = (random.random() * transition_behaviour.transition_coefficient(
            current_time, self.drift_time,
            self.drift_radius
        ))
        flag = round(prob)
        coefficient_delta = flag * self.coefficient_shift
        return coefficient_delta

    @property
    def _drift_name(self) -> str:
        """
        The name of the drift behavior.

        Returns
        -------
        str
            The name of the drift behavior: "gradual".
        """
        return "gradual"



import numpy as np
import random
from sid2re.driftgenerator.concept.nodes.root_distributions import UniformDistribution, PeriodicalDistribution, \
    ConstantDistribution, GaussianDistribution


def test_distributions_faulty_sensor_behaviour():
    for dist_class in [UniformDistribution, PeriodicalDistribution, ConstantDistribution, GaussianDistribution]:
        value = ConstantDistribution.generate(w=np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]), timestamp=0,
                                              minimum=0, maximum=1)
        assert value == 0.0


def test_constant():
    random.seed(42)
    np.random.seed(42)
    value = ConstantDistribution.generate(w=np.array([0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0]), timestamp=0,
                                          minimum=0, maximum=1)
    assert value == 0.5
    value = ConstantDistribution.generate(w=np.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0]), timestamp=0,
                                          minimum=1, maximum=1)
    assert value == 1.0
    value = ConstantDistribution.generate(w=np.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0]), timestamp=0,
                                          minimum=-1, maximum=-1)
    assert value == -1.0
    value = ConstantDistribution.generate(w=np.array([0.5, 0, 0, 0, 0, 0, 0, 0, 0, 0]), timestamp=0,
                                          minimum=-1, maximum=1)
    assert value == 0.0


def test_periodical():
    random.seed(42)
    np.random.seed(42)
    epsilon = 0.1
    value = [PeriodicalDistribution.generate(w=np.array([0.1, 0.5, 1, 0.5, 1, 0, 0, 0, 0, 0]),
                                             timestamp=i, minimum=-5, maximum=5) for i in np.arange(10000)]
    assert (0.0 - epsilon <= np.mean(value) <= 0.0 + epsilon)
    assert (np.amax(value) <= 5.0)
    assert (-5.0 <= np.amin(value))
    value = [PeriodicalDistribution.generate(w=np.array([0.4, 0, 1, 0, 1, 0, 0, 0, 0, 0]),
                                             timestamp=i, minimum=-5, maximum=5) for i in np.arange(10000)]
    assert (0.0 - epsilon <= np.mean(value) <= 0.0 + epsilon)
    assert (np.amax(value) <= 5.0)
    assert (-5.0 <= np.amin(value))
    value = [PeriodicalDistribution.generate(w=np.array([0.7, 0, 1, 0, 1, 0, 0, 0, 0, 0]),
                                             timestamp=i, minimum=-5, maximum=5) for i in np.arange(10000)]
    print(np.mean(value))
    assert (0.0 <= np.mean(value))
    assert (np.amax(value) <= 5.0)
    assert (-5.0 <= np.amin(value))
    value = [PeriodicalDistribution.generate(w=np.array([0.7, 0, 1, 0, 1, 1, 0, 0, 0, 0]),
                                             timestamp=i, minimum=-5, maximum=5) for i in np.arange(10000)]
    print(np.mean(value))
    assert (np.mean(value) <= 0.0)
    assert (np.amax(value) <= 5.0)
    assert (-5.0 <= np.amin(value))


def test_gaussian():
    random.seed(42)
    np.random.seed(42)
    epsilon = 0.1
    value = [GaussianDistribution.generate(w=np.array([0.6, 0.1, 0, 0, 0, 0, 0, 0, 0, 0]),
                                           timestamp=i, minimum=-5, maximum=5) for i in np.arange(10000)]
    assert (1.0 - epsilon <= np.mean(value) <= 1.0 + epsilon)
    assert (0.1 - epsilon <= np.var(value) <= 0.1 + epsilon)
    assert (np.amax(value) <= 5.0)
    assert (-5.0 <= np.amin(value))
    value = [GaussianDistribution.generate(w=np.array([0.6, 10, 0, 0, 0, 0, 0, 0, 0, 0]),
                                           timestamp=i, minimum=-5, maximum=5) for i in np.arange(10000)]
    assert (np.amax(value) <= 5.0)
    assert (-5.0 <= np.amin(value))


def test_uniform():
    random.seed(42)
    np.random.seed(42)
    epsilon = 0.1
    value = [UniformDistribution.generate(w=np.array([0.6, 0.1, 0, 0, 0, 0, 0, 0, 0, 0]),
                                          timestamp=i, minimum=-5, maximum=5) for i in np.arange(10000)]
    assert (1.0 - epsilon <= np.mean(value) <= 1.0 + epsilon)
    assert (np.amax(value) <= 5.0)
    assert (-5.0 <= np.amin(value))



from sid2re.driftgenerator.utils.metrics import error, ae, mae, mse, rmse, mape, bias, drift_correction_error, \
    drift_reference_error
import numpy as np
import pytest


def test_formatting_no_averaging():
    formats = [(1, 1), (2, 1), (10, 1), (10, 2), (10, 5)]
    for form in formats:
        pred = np.random.rand(*form)
        truth = np.random.rand(*form)
        ref = np.random.rand(*form)
        for met in [error, ae]:
            print(met)
            res = met(pred, truth)
            assert res.shape == form
        for met in [mae, mse, rmse, mape, bias, drift_correction_error]:
            print(met)
            res = met(pred, truth)
            assert len(res.shape) == 1
            assert res.shape[0] == form[1]
        for met in [drift_reference_error]:
            print(met)
            res = met(pred, truth, ref)
            assert len(res.shape) == 1
            assert res.shape[0] == form[1]


@pytest.mark.parametrize('metric', [error, ae, mae, mse, rmse, mape, bias, drift_correction_error])
def test_metrics_enforce_type_error(metric):
    with pytest.raises(ValueError) as excinfo:
        _ = metric([1, 2, 3, 4], [5, 6, 7, 8])
    assert len(str(excinfo.value)) > 0


def test_formatting_averaging():
    formats = [(1, 1), (2, 1), (10, 1), (10, 2), (10, 5)]
    for form in formats:
        pred = np.random.rand(*form)
        truth = np.random.rand(*form)
        ref = np.random.rand(*form)
        for met in [mae, mse, rmse, mape, bias, drift_correction_error]:
            print(met)
            res = met(pred, truth, averaging=True)
            assert isinstance(res, float)
        for met in [drift_reference_error]:
            print(met)
            res = met(pred, truth, ref, averaging=True)
            assert isinstance(res, float)



import numpy as np

from sid2re.driftgenerator.concept.nodes._base_node import _BaseNode
from sid2re.driftgenerator.concept.nodes.root_distributions import _BaseDistribution, UniformDistribution


class RootFeatureNode(_BaseNode):
    """Noe representing an independent feature. Also called root feature."""

    def __init__(self, minimum: float, maximum: float):
        """
        Initializes a RootFeatureNode instance.

        Parameters
        ----------
        minimum : float
            The minimum value of the generated feature.
        maximum : float
            The maximum value of the generated feature.
        """
        super().__init__()
        self.minimum: float = minimum
        self.maximum: float = maximum
        self.distribution: _BaseDistribution = UniformDistribution()

    def set_distro(self, distribution: _BaseDistribution) -> None:
        """
        Sets the distribution type.

        Parameters
        ----------
        distribution : _BaseDistribution
            The distribution type.
        """
        self.distribution = distribution

    def generate_data(self, time_stamps: np.ndarray) -> np.ndarray:
        """
        Generates data based on the configured distribution.

        Parameters
        ----------
        time_stamps : np.ndarray
            List of time stamps.

        Returns
        -------
        np.ndarray
            List of generated data.
        """

        return np.array(
            [self.distribution.generate(self.current_config(time), time, self.minimum, self.maximum) for time in
             time_stamps]
        )



from sid2re.driftgenerator.concept.drift_transition_functions._base_transition_function import _BaseTransitionFunction
from sid2re.driftgenerator.concept.drift_transition_functions.linear_transition_function import LinearTransitionFunction

__all__ = ["_BaseTransitionFunction", "LinearTransitionFunction"]



from sid2re.driftgenerator.concept.drift_transition_functions._base_transition_function import _BaseTransitionFunction


class LinearTransitionFunction(_BaseTransitionFunction):
    """Transition function used by incremental drifts. This is a linear transition."""
    @classmethod
    def _transition_coefficient(cls, current_time: float, shift_centre: float, shift_radius: float) -> float:
        """Compute the transition coefficient using linear interpolation.

        Parameters
        ----------
        current_time : float
            The current time in the system.
        shift_centre : float
            The center of the shift.
        shift_radius : float
            The radius around the shift center.

        Returns
        -------
        float
            The computed transition coefficient.
        """
        drift_begin = shift_centre - shift_radius
        drift_end = shift_centre + shift_radius
        drift_interval = drift_end - drift_begin
        passed_interval = current_time - drift_begin

        return passed_interval / drift_interval



import random

from sid2re.driftgenerator.concept.nodes.root_distributions._base_distribution import _BaseDistribution
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class GaussianDistribution(_BaseDistribution):
    """Value generating distribution used in root features.

    This distribution generates random values based on a gaussian distribution.
    """

    @classmethod
    def _generate(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:  WPS123
        """
        Generate a random value using a Gaussian distribution.

        Parameters
        ----------
        w : NumberArray
            Behaviour defining vector of values between 0 and 1.
            w[0]: centre of the uniform distribution.
            w[1]: share of [minimum, maximum] that is used as variance for the Gaussian distribution.
        timestamp : float
            Point in time for which the function value of the function is asked (not used in this case).
        minimum : float
            Minimal function value that is allowed to be generated.
        maximum : float
            Maximal function value that is allowed to be generated.

        Returns
        -------
        float
            The random value sampled from the Gaussian distribution defined by w.
        """
        generated_value = random.gauss(w[0] * (maximum - minimum) + minimum, w[1] * (maximum - minimum) / 10)
        if generated_value > maximum:
            generated_value = maximum
        elif generated_value < minimum:
            generated_value = minimum
        return float(generated_value)



import random

from sid2re.driftgenerator.concept.nodes.root_distributions._base_distribution import _BaseDistribution
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class UniformDistribution(_BaseDistribution):
    """Value generating distribution used in root features.

    This distribution randomly generates values based on a uniform distribution.
    """

    @classmethod
    def _generate(cls, w: NumberArray, timestamp: float, minimum: float, maximum: float) -> float:  # noqa:  WPS123
        """Generate a value based on a uniform distribution.

        Parameters
        ----------
        w : NumberArray
            Behaviour defining vector of values between 0 and 1.
            w[0]: share of the interval used for shifting the function along the time axis.
            w[1]: defining periodical interval between 5 and 50.
            w[2]: share of the [feature_min, feature_max] interval, used for value axis shift.
            w[3]: proportional to pitch of the linear function.
            w[4]: used to determine whether the pitch of the linear function is positive or negative.
        timestamp : float
            Point in time for which the function value of the transformed linear function is asked.
        minimum : float
            Minimal function value that is allowed to be generated.
        maximum : float
            Maximal function value that is allowed to be generated.

        Returns
        -------
        float
            The generated value.
        """
        generated_value = random.random() * w[1] * (maximum - minimum)
        max_val = w[1] * (maximum - minimum)
        centered_val = generated_value - max_val / 2
        offset = (w[0] * ((maximum - max_val / 2) - (minimum + max_val / 2))) + (minimum + max_val / 2)
        return float(centered_val + offset)



from pathlib import Path


# Seaborn visualization library


def get_script_location() -> Path:
    """
    Gets the root directory of the project by resolving the known path relative to this file.

    Returns
    -------
    Path
        The project root directory.
    """

    return Path(__file__)



from sid2re.driftgenerator.concept.drift_behaviours._base_drift_behaviour import _BaseDriftBehaviour
from sid2re.driftgenerator.concept.drift_transition_functions import _BaseTransitionFunction
from sid2re.driftgenerator.utils.type_aliases import NumberArray


class SuddenDriftBehaviour(_BaseDriftBehaviour):
    """Specify the transformation of the concept coefficients caused by a sudden drift.

    A sudden drift instantly implements a new concept.
    """
    def _compute_coeff_delta(
        self,
        coefficient: NumberArray,
        transition_behaviour: _BaseTransitionFunction,
        current_time: float
    ) -> NumberArray:
        """Compute the change in coefficients during sudden drift.

        Parameters
        ----------
        coefficient : NumberArray
            The coefficients to be modified.
        transition_behaviour : _BaseTransitionFunction
            The behavior of transition.
        current_time : float
            The current time in the system.

        Returns
        -------
        NumberArray
            The computed change in coefficients.
        """
        coefficient_delta = self.coefficient_shift
        return coefficient_delta

    @property
    def _drift_name(self) -> str:
        """The name of the drift behavior.

        Returns
        -------
        str
            The name of the drift behavior: "sudden".
        """
        return "sudden"
